"""A script containing helper methods used throughout the package.
"""

import sys
import os
import re
from importlib.util import find_spec
import inspect
from typing import Optional, Any, Dict, Callable, Union
from io import IOBase, BytesIO
from uuid import uuid4
from mimetypes import guess_extension
from zipfile import ZipFile
import requests

def get_default(func: Callable[..., Any], param: str) -> Optional[Any]:
    """Gets the default value of a function parameter, or `None` if not applicable.
    
    Parameters
    ----------
    func : Callable[..., Any]
        The function to check.
    param : str
        The name of the parameter.

    Returns
    -------
    Any | None
        The default value of the parameter, or `None` if not applicable.
    """
    param_sig: inspect.Parameter = inspect.signature(func).parameters[param]
    return None if param_sig.default is inspect.Parameter.empty else param_sig.default

def get_or_default(dict_obj: Dict[str, Any], key: str, func: Callable[..., Any],
        param: Optional[str] = None) -> Optional[Any]:
    """Gets the value within a dictionary, or the default from the function if none is specified.

    Parameters
    ----------
    dict_obj : Dict[str, Any]
        The dictionary containing the data of an object.
    key : str
        The key to obtain the value of the dictionary from.
    func : Callable[..., Any]
        The function to check the default of if the key does not exist in the dictionary.
    param : str | None (default None)
        The name of the parameter containing the default value. When `None`, defaults to the `key`.
    
    Returns
    -------
    Any | None
        The value of the key, the default value, or `None`.
    """
    if param is None:
        param: str = key

    return dict_obj[key] if key in dict_obj else get_default(func, param)

def has_module(name: str) -> bool:
    """Checks whether the module is currently loaded or can be added to the current workspace.

    Parameters
    ----------
    name : str
        The name of the module.

    Returns
    -------
    bool
        `True` if the module exists, `False` otherwise.
    """
    return (name in sys.modules) or (find_spec(name) is not None)

def unzip(file: Union[str, IOBase], out_dir: str = os.curdir) -> None:
    """Unzips the file or stream to the specified directory.

    Parameters
    ----------
    file : str | io.IOBase
        The file or stream of the zip file.
    out_dir : str (default '.')
        The directory to extract the zip file to.
    """
    with ZipFile(file, 'r') as zip_ref: # type: ZipFile
        zip_ref.extractall(out_dir)

_FILENAME_REGEX: str = \
    r'filename\*?=(?:\"([^\'\"\n;]+)\"|(?:(?:UTF-8|ISO-8859-1|[^\'\"])\'[^\'\"]?\')([^\'\"\n;]+));?'
"""Regex for getting the filename from the content-disposition header.
Tries to read normal filename and a fuzzy regex of the
[RFC8187](https://datatracker.ietf.org/doc/html/rfc8187) spec.
"""

_CONTENT_DISPOSITION: str = 'content-disposition'
"""The header for the content disposition."""

_CONTENT_TYPE: str = 'content-type'
"""The header for the content type."""

def download_file(url: str, handler: Callable[[requests.Response, str], bool],
        stream: bool = True) -> bool:
    """Downloads a file from the specified url via a GET request and handles the response
    bytes as specified.
    
    Parameters
    ----------
    url : str
        The url to download the file from.
    handler : (requests.Response, str) -> bool
        A function which takes in the response and filename and returns whether the file was
        successfully handled.
    stream : bool (default True)
        If `False`, the response content will be immediately downloaded.

    Returns
    -------
    bool
        `True` if the file was successfully downloaded, `False` otherwise
    """

    # Download data within 5 minutes
    with requests.get(url, stream = stream, allow_redirects = True,
            timeout = 300) as response: # type: requests.Response
        # If cannot grab file, return False
        if not response.ok:
            return False

        # Get filename
        filename: Optional[str] = None

        ## Lookup filename from content disposition if present
        if _CONTENT_DISPOSITION in response.headers:
            for filename_lookup in re.findall(_FILENAME_REGEX,
                    response.headers[_CONTENT_DISPOSITION]): # type: Tuple[str, str]
                # If filename* is present, set and then break
                if (name := filename_lookup[1]): # name: str
                    filename = name
                    break
                # Otherwise, set the normal filename and keep checking
                filename = filename_lookup[0]

        ## If no filename was present, assign a default name
        if filename is None:
            filename: str = str(uuid4())
            # Set file extension from content type, if available
            if _CONTENT_TYPE in response.headers:
                if (ext := guess_extension(
                        response.headers[_CONTENT_TYPE].partition(';')[0].strip()
                    )) is not None: # ext: Optional[str]
                    filename += ext

        # Handle the result of the downloaded file
        return handler(response, filename)

def download_and_write(url: str, unzip_file: bool = True, out_dir: str = os.curdir,
        stream: bool = True) -> bool:
    """Downloads a file from the specified url via a GET request and writes or unzips
    the file, if applicable.

    Parameters
    ----------
    url : str
        The url to download the file from.
    unzip_file : bool (default True)
        If `True`, will attempt to unzip the file if the file extension is correct.
    out_dir : str (default '.')
        The directory to write or unzip the file to.
    stream : bool (default True)
        If `False`, the response content will be immediately downloaded.

    Returns
    -------
    bool
        `True` if the file was successfully downloaded, `False` otherwise
    """

    def __write(__response: requests.Response, __filename: str, __dir: str) -> bool:
        """Writes the file or unzips it to the specified directory.

        Parameters
        ----------
        __response : requests.Response
            The response of the url request.
        __filename : str
            The name of the file requested from the url.
        __dir : str
            The directory to write the file(s) to.
        
        Returns
        -------
        bool
            `True` if the download was successful, `False` otherwise.
        """

        # Unzip file if available and set
        if unzip_file and __filename.endswith('.zip'):
            with BytesIO(__response.content) as zip_bytes: # type: BytesIO
                unzip(zip_bytes, out_dir = __dir)
        # Otherwise do normal extraction
        else:
            # Create directory name if not already present
            name: str = os.sep.join([__dir, __filename])
            os.makedirs(os.path.dirname(name), exist_ok = True)

            with open(name, 'wb') as file:
                for data in __response.iter_content(1024): # type: ReadableBuffer
                    file.write(data)
        return True

    return download_file(url,
        lambda response, filename: __write(response, filename, out_dir), stream = stream)



"""A script obtaining a project file from a download url.
"""

import os
from typing import Optional
from project_patcher.lazy import SINGLETON
from project_patcher.struct.codec import DictObject
from project_patcher.metadata.file import ProjectFile, ProjectFileCodec, build_file
from project_patcher.utils import download_and_write

class URLProjectFile(ProjectFile):
    """A project file for a file at a downloadable url link.
    The file will be obtained via a GET request."""

    def __init__(self, url: str, rel_dir: str = os.curdir,
            extra: Optional[DictObject] = None) -> None:
        """
        Parameters
        ----------
        url : str
            The downloadable link for the file.
        rel_dir : str (default '.')
            The directory the project file is located.
        extra : dict[str, Any]
            Extra data defined by the user.
        """
        super().__init__(rel_dir, extra)
        self.url: str = url

    def codec(self) -> 'ProjectFileCodec':
        return SINGLETON.URL_FILE_CODEC

    def setup(self, root_dir: str) -> bool:
        super().setup(root_dir)
        return download_and_write(self.url, unzip_file = False,
            out_dir = self._create_path(root_dir))

def build_url() -> URLProjectFile:
    """Builds an URLProjectFile from user input.
    
    Returns
    -------
    URLProjectFile
        The built project file.
    """
    url: str = input('Direct URL Link: ')
    return build_file(lambda rel_dir: URLProjectFile(url, rel_dir = rel_dir))

class URLProjectFileCodec(ProjectFileCodec[URLProjectFile]):
    """A codec for encoding and decoding an URLProjectFile.
    """

    def encode_type(self, obj: URLProjectFile, dict_obj: DictObject) -> DictObject:
        dict_obj['url'] = obj.url
        return dict_obj

    def decode_type(self, rel_dir: str, extra: Optional[DictObject],
            obj: DictObject) -> URLProjectFile:
        return URLProjectFile(obj['url'], rel_dir = rel_dir, extra = extra)



"""A script containing a generic implementation of a registry system.
"""

from typing import TypeVar, Dict, Generic

RegistryObject = TypeVar('RegistryObject')
"""The type of the values in a registry."""

class Registry(Dict[str, RegistryObject], Generic[RegistryObject]):
    """A string key to value bidirectional dictionary.
    
    Types
    -----
    RegistryObject
        The type of the values in the registry.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__inverse: Dict[RegistryObject, str] = {}

    def __setitem__(self, __key: str, __value: RegistryObject) -> None:
        # Make sure the value isn't already assigned to a given key
        if __value in self.__inverse:
            raise ValueError(f'{__value} cannot be assigned to {__key} \
                as it already exists for {self.get_key(__value)}.')

        super().__setitem__(__key, __value)
        self.__inverse[__value] = __key

    def __delitem__(self, __key: str) -> None:
        del self.__inverse[self[__key]]
        super().__delitem__(__key)

    def get_key(self, __value: RegistryObject) -> str:
        """Gets the key from the value.

        Parameters
        ----------
        __value : RegistryObject
            The value to get the key of.

        Return
        ------
        str
            The key of the value.
        """
        return self.__inverse[__value]



"""A script obtaining a project file from an Open Science Framework
project.
"""

import os
from typing import Optional
from project_patcher.lazy import SINGLETON
from project_patcher.struct.codec import DictObject
from project_patcher.metadata.file import ProjectFile, ProjectFileCodec, build_file
from project_patcher.utils import download_and_write

class OSFProjectFile(ProjectFile):
    """A project file for an Open Science Framework repository."""

    def __init__(self, project_id: str, rel_dir: str = os.curdir,
            extra: Optional[DictObject] = None) -> None:
        """
        Parameters
        ----------
        project_id : str
            The five character identifier of the repository.
        rel_dir : str (default '.')
            The directory the project file is located.
        extra : dict[str, Any]
            Extra data defined by the user.
        """
        super().__init__(rel_dir, extra)
        self.project_id: str = project_id
        self.__url: str = \
            f'https://files.osf.io/v1/resources/{project_id}/providers/osfstorage/?zip='

    def codec(self) -> 'ProjectFileCodec':
        return SINGLETON.OSF_FILE_CODEC

    def setup(self, root_dir: str) -> bool:
        super().setup(root_dir)
        return download_and_write(self.__url, out_dir = self._create_path(root_dir))

def build_osf() -> OSFProjectFile:
    """Builds an OSFProjectFile from user input.
    
    Returns
    -------
    OSFProjectFile
        The built project file.
    """
    project_id: str = input('OSF Project Id: ')
    return build_file(lambda rel_dir: OSFProjectFile(project_id, rel_dir = rel_dir))

class OSFProjectFileCodec(ProjectFileCodec[OSFProjectFile]):
    """A codec for encoding and decoding an OSFProjectFile.
    """

    def encode_type(self, obj: OSFProjectFile, dict_obj: DictObject) -> DictObject:
        dict_obj['id'] = obj.project_id
        return dict_obj

    def decode_type(self, rel_dir: str, extra: Optional[DictObject],
            obj: DictObject) -> OSFProjectFile:
        return OSFProjectFile(obj['id'], rel_dir = rel_dir, extra = extra)



"""A script which holds all lazy references to a given
class."""

import sys
from types import ModuleType
from typing import Optional, Dict
from importlib.machinery import ModuleSpec
from importlib.util import find_spec, LazyLoader, module_from_spec

_LAZY_IMPORTS: Dict[str, ModuleType] = {}
"""A dictionary containing all the modules that were lazily imported."""

def _lazy_import(name: str) -> ModuleType:
    """Sets up the module to be lazily imported if it is not already loaded.

    Parameters
    ----------
    name : str
        The name of the module to lazily load.
    
    Returns
    -------
    types.ModuleType
        The module which will be loaded on first execution.
    """

    # If already loaded, just return the module
    if name in sys.modules:
        return sys.modules[name]

    # Otherwise, if the import is already lazy, return the module
    if name in _LAZY_IMPORTS:
        return _LAZY_IMPORTS[name]

    # Otherwise, make import lazy and return module
    spec: Optional[ModuleSpec] = find_spec(name)
    loader: LazyLoader = LazyLoader(spec.loader)
    spec.loader = loader
    module: ModuleType = module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    _LAZY_IMPORTS[name] = module
    return module

SINGLETON: ModuleType = _lazy_import('project_patcher.singleton')
"""The project_patcher.singleton module added as a lazy import."""



"""A script containing information about handling the project metadata.
"""

import os
from typing import List, Optional, Tuple, Set
from pathlib import Path
import fnmatch
from project_patcher.lazy import SINGLETON
from project_patcher.metadata.file import ProjectFile
from project_patcher.struct.codec import DictCodec, DictObject
from project_patcher.utils import get_or_default

class ProjectMetadata:
    """Metadata information associated with the project being patched or ran."""

    def __init__(self, files: List[ProjectFile],
            ignore: Optional[List[str]] = None, overwrite: Optional[List[str]] = None,
            extra: Optional[DictObject] = None) -> None:
        """
        Parameters
        ----------
        files : list of `ProjectFile`s
            The files associated with the project.
        ignore : list of str
            The patterns for files that are ignored for patching.
        overwrite: list of str
            The patterns for files that are overwritten instead of patching.
        extra : dict[str, Any]
            Extra data defined by the user.
        """
        self.files: List[ProjectFile] = files
        self.ignore: List[str] = [] if ignore is None else ignore
        self.overwrite: List[str] = [] if overwrite is None else overwrite
        self.extra: DictObject = {} if extra is None else extra

    def setup(self, root_dir: str) -> bool:
        """Sets up the project for usage.

        Parameters
        ----------
        root_dir : str
            The root directory to set up the project in.
        """
        failed: List[ProjectFile] = []

        for file in self.files: # type: ProjectFile
            if not file.setup(root_dir):
                failed.append(file)

        return not failed

    def codec(self) -> 'ProjectMetadataCodec':
        """Returns the codec used to encode and decode this metadata.

        Returns
        -------
        ProjectFileCodec
            The codec used to encode and decode this metadata.
        """
        return SINGLETON.METADATA_CODEC

    def ignore_and_overwrite(self, root_dir: str) -> Tuple[Set[str], Set[str]]:
        """Gets the ignored and overwritten files from the specified directory.

        Parameters
        ----------
        root_dir : str
            The root directory to get the files of.
        
        Returns
        -------
        (set of strs, set of strs)
            A tuple of ignored and overwritten files, respectively.
        """

        # Get all files
        files: list[str] = [Path(os.path.relpath(path, root_dir)).as_posix() \
            for path in Path(root_dir).rglob('*') if path.is_file()]

        # Get ignored files
        ignored_files: Set[str] = set()
        for ignore in self.ignore: # type: str
            ignored_files.update(fnmatch.filter(files, ignore))

        # Get overwritten files
        overwritten_files: Set[str] = set()
        for overwrite in self.overwrite: # type: str
            overwritten_files.update(fnmatch.filter(files, overwrite))

        return (ignored_files, overwritten_files)

def build_metadata() -> ProjectMetadata:
    """Builds a ProjectMetadata from user input.
    
    Returns
    -------
    ProjectMetadata
        The built project metadata.
    """

    # Project Files
    available_file_types: str = ', '.join(SINGLETON.PROJECT_FILE_BUILDERS.keys())
    files: List[ProjectFile] = []
    flag: bool = True
    while flag:
        file_type: str = input(f"Add file ({available_file_types}): ").lower()
        files.append(SINGLETON.PROJECT_FILE_BUILDERS[file_type]())
        flag = input('Would you like to add another file (y/n)? ').lower()[0] != 'n'

    # Ignore Patterns
    flag = input('Would you like to ignore any files when patching (y/n)? ').lower()[0] != 'n'
    ignore: List[str] = []
    while flag:
        ignore.append(input('Add pattern to ignore: '))
        flag = input('Would you like to ignore another pattern (y/n)? ').lower()[0] != 'n'

    # Overwrite Patterns
    flag = input('Would you like to overwrite any files when patching (y/n)? ').lower()[0] != 'n'
    overwrite: List[str] = []
    while flag:
        overwrite.append(input('Add pattern to overwrite: '))
        flag = input('Would you like to overwrite another pattern (y/n)? ').lower()[0] != 'n'

    # Create metadata
    return ProjectMetadata(files, ignore = ignore, overwrite = overwrite)

class ProjectMetadataCodec(DictCodec[ProjectMetadata]):
    """A codec for encoding and decoding a ProjectMetadata.
    """

    def encode(self, obj: ProjectMetadata) -> DictObject:
        dict_obj: DictObject = {}
        dict_obj['files'] = list(map(lambda file: file.codec().encode(file), obj.files))
        if obj.ignore:
            dict_obj['ignore'] = obj.ignore
        if obj.overwrite:
            dict_obj['overwrite'] = obj.overwrite
        if obj.extra:
            dict_obj['extra'] = obj.extra
        return dict_obj

    def __decode_file(self, file: DictObject) -> ProjectFile:
        """Decodes a project file from its type.

        Parameters
        ----------
        file : Dict[str, Any]
            The encoded project file.
        
        Returns
        -------
        `ProjectFile`
            The decoded project file.
        """
        return SINGLETON.PROJECT_FILE_TYPES[file['type']].decode(file)

    def decode(self, obj: DictObject) -> ProjectMetadata:
        return ProjectMetadata(list(map(self.__decode_file, obj['files'])),
            ignore = get_or_default(obj, 'ignore', ProjectMetadata),
            overwrite = get_or_default(obj, 'overwrite', ProjectMetadata),
            extra = get_or_default(obj, 'extra', ProjectMetadata))



"""A script containing information about handing a project file read
from the metadata.
"""

import os
from typing import TypeVar, List, Callable, Optional
from abc import ABC, abstractmethod
from project_patcher.lazy import SINGLETON
from project_patcher.utils import get_default, get_or_default
from project_patcher.struct.codec import DictObject, DictCodec

class ProjectFile(ABC):
    """An abstract class containing information about a file associated with the project.
    """

    @abstractmethod
    def __init__(self, rel_dir: str = os.curdir,
            extra: Optional[DictObject] = None) -> None:
        """
        Parameters
        ----------
        rel_dir : str (default '.')
            The directory the project file is located.
        extra : dict[str, Any]
            Extra data defined by the user.
        """
        super().__init__()
        self.dir: str = rel_dir
        self.extra: DictObject = {} if extra is None else extra

    @abstractmethod
    def codec(self) -> 'ProjectFileCodec':
        """Returns the codec used to encode and decode this project file.

        Returns
        -------
        ProjectFileCodec
            The codec used to encode and decode this project file.
        """

    @abstractmethod
    def setup(self, root_dir: str) -> bool:
        """Sets up the project file for usage.

        Parameters
        ----------
        root_dir : str
            The root directory to set up the project file in.
        """
        os.makedirs(self._create_path(root_dir), exist_ok = True)
        return False

    def _create_path(self, root_dir: str, *paths: str) -> str:
        """Constructs a path from the root directory through the relative
        directory and any additional paths specified.

        Parameters
        ----------
        root_dir : str
            The root directory to create the path from.
        *paths : str
            The paths after the project file's relative directory.

        Returns
        -------
        str
            The newly created path.
        """
        fpath: List[str] = [root_dir, self.dir]
        fpath += paths
        return os.sep.join(fpath)

PF = TypeVar('PF', bound = ProjectFile)
"""The type of the project file."""

def build_file(callback: Callable[[str], PF]) -> PF:
    """Builds a ProjectFile from user input based on the
    passed in callback.
    
    Parameters
    ----------
    callback : (str) -> ProjectFile
        A function that takes in the root directory of the ProjectFile
        and returns the completed ProjectFile.
    
    Returns
    -------
    ProjectFile
        The built project file.
    """
    def_rel_dir: str = get_default(ProjectFile, 'rel_dir')
    rel_dir: str = \
        input(f"Directory to extract to (default '{def_rel_dir}'): ")
    return callback(rel_dir if rel_dir else def_rel_dir)

class ProjectFileCodec(DictCodec[PF]):
    """An abstract, generic encoder and decoder between a dictionary and a ProjectFile.

    Types
    -----
    PF
        The type of the project file to be encoded or decoded to.
    """

    def decode(self, obj: DictObject) -> PF:
        return self.decode_type(get_or_default(obj, 'dir', ProjectFile, param = 'rel_dir'),
            get_or_default(obj, 'extra', ProjectFile), obj)

    @abstractmethod
    def decode_type(self, rel_dir: str, extra: Optional[DictObject], obj: DictObject) -> PF:
        """Decodes a dictionary to the specific ProjectFile type.

        Parameters
        ----------
        rel_dir : str
            The directory the ProjectFile is located.
        obj : Dict[str, Any]
            The dictionary containing the data for the ProjectFile.

        Returns
        -------
        PF
            The decoded ProjectFile.
        """

    def encode(self, obj: PF) -> DictObject:
        dict_obj: DictObject = {}
        dict_obj['type'] = SINGLETON.PROJECT_FILE_TYPES.get_key(self)
        if obj.dir != get_default(ProjectFile, 'rel_dir'):
            dict_obj['dir'] = obj.dir
        if obj.extra:
            dict_obj['extra'] = obj.extra

        return self.encode_type(obj, dict_obj)

    @abstractmethod
    def encode_type(self, obj: PF, dict_obj: DictObject) -> DictObject:
        """Encodes a specific ProjectFile type to the dictionary.

        Parameters
        ----------
        obj : PF
            The ProjectFile containing the data.
        dict_obj : Dict[str, Any]
            The dictionary containing some common encoded data.

        Returns
        -------
        Dict[str, Any]
            The encoded ProjectFile in a dictionary.
        """



"""A script obtaining a project file from a git repository.
"""

import os
from typing import Set, Optional
from git import Repo
from project_patcher.lazy import SINGLETON
from project_patcher.utils import get_default
from project_patcher.struct.codec import DictObject
from project_patcher.metadata.file import ProjectFile, ProjectFileCodec, build_file

_VALID_BRANCH_TYPES: Set[str] = {
    'branch',
    'commit',
    'tag'
}
"""A set of valid keys indicating the checkout location of the Git repository."""

class GitProjectFile(ProjectFile):
    """A project file for an Git repository."""

    def __init__(self, repository: str, branch: Optional[str] = None,
            branch_type: str = 'branch', rel_dir: str = os.curdir,
            extra: Optional[DictObject] = None) -> None:
        """
        Parameters
        ----------
        repository : str
            The Git link for the repository location.
        branch : str | None (default None)
            The name of the checkout location. If `None`, the default checkout
            location will be used.
        branch_type : str (default 'branch')
            The name of the key holding the branch. Must be within `_VALID_BRANCH_TYPES`.
        rel_dir : str (default '.')
            The directory the project file is located.
        extra : dict[str, Any]
            Extra data defined by the user.
        """
        super().__init__(rel_dir, extra)
        self.repository: str = repository
        self.branch: Optional[str] = branch
        if branch_type not in _VALID_BRANCH_TYPES:
            raise ValueError(f'\"{branch_type}\" is not a valid branch type. \
                Specify one of the following: {", ".join(_VALID_BRANCH_TYPES)}')
        self.branch_type: str = branch_type

    def codec(self) -> 'ProjectFileCodec':
        return SINGLETON.GIT_FILE_CODEC

    def setup(self, root_dir: str) -> bool:
        super().setup(root_dir)

        # Checkout and change branches, if applicable
        repo: Repo = Repo.clone_from(self.repository, self._create_path(root_dir))
        if self.branch is not None:
            repo.git.checkout(self.branch)
        return True

def build_git() -> GitProjectFile:
    """Builds a GitProjectFile from user input.
    
    Returns
    -------
    GitProjectFile
        The built project file.
    """
    repository: str = input('Git Repository: ')
    def_branch_type: str = get_default(GitProjectFile, 'branch')
    branch_type: str = \
        input(f"{', '.join(_VALID_BRANCH_TYPES)} (default '{def_branch_type}'): ").lower()
    branch: Optional[str] = input(f"{branch_type} (default branch): ")
    if not branch:
        branch = None
    return build_file(lambda rel_dir:
        GitProjectFile(repository, branch = branch, branch_type = branch_type, rel_dir = rel_dir)
    )

class GitProjectFileCodec(ProjectFileCodec[GitProjectFile]):
    """A codec for encoding and decoding a GitProjectFile.
    """

    def encode_type(self, obj: GitProjectFile, dict_obj: DictObject) -> DictObject:
        dict_obj['repository'] = obj.repository
        if obj.branch != get_default(GitProjectFile, 'branch'):
            dict_obj[obj.branch_type] = obj.branch
        return dict_obj

    def decode_type(self, rel_dir: str, extra: Optional[DictObject],
            obj: DictObject) -> GitProjectFile:
        for branch_name in _VALID_BRANCH_TYPES: # type: str
            if branch_name in obj:
                return GitProjectFile(
                    obj['repository'],
                    branch = obj[branch_name],
                    branch_type = branch_name, rel_dir = rel_dir
                )
        return GitProjectFile(obj['repository'], rel_dir = rel_dir, extra = extra)



"""A script containing the methods needed to create and apply patches
between two blocks of text.
"""

# License: Public domain (CC0)
# Isaac Turner 2016/12/05

import re
from difflib import unified_diff
from typing import Iterator, List, Optional
from datetime import datetime

_NO_EOL: str = '\\ No newline at end of file'
"""Text indicating there is no newline at the end of the diff file."""

_HUNK_HEADER: re.Pattern = re.compile(r'^@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@$')
"""The regex to get the file line information from the hunk header."""

class PatchError(ValueError):
    """The patch file contained invalid information.
    """

def create_patch(from_text: str, to_text: str, filename: str = '',
        time: str = str(datetime.now())) -> str:
    """Creates a patch between two pieces of text. If equivalent, returns
    an empty string.

    Parameters
    ----------
    from_text : str
        The original text to generate the patch from.
    to_text : str
        The new text the patch will transform the original text to.
    filename : str (default '')
        The name of the file the patch is being applied for.
    time : datetime.datetime (default str(datetime.datetime.now))
        The time the patch was generated.

    Returns
    -------
    str
        The generated patch.
    """

    diffs: Iterator[str] = unified_diff(
        from_text.splitlines(keepends = True),
        to_text.splitlines(keepends = True),
        fromfile = filename, tofile = filename,
        tofiledate = time
    )
    # Join diffs together
    # Apply no eol if text line doesn't end with a newline
    return ''.join([diff if diff[-1] == '\n' else f'{diff}\n{_NO_EOL}\n' for diff in diffs])

def apply_patch(text: str, patch: str, revert: bool = False) -> str:
    """Applies a patch to the current text.

    Parameters
    ----------
    text : str
        The text to apply the patch to.
    patch : str
        The patch to apply.
    revert : bool (default False)
        When `True`, will attempt to recover the original file.

    Returns
    -------
    str
        The patched text.
    """

    # Setup basic references and result vars
    text: List[str] = text.splitlines(keepends = True)
    patch: List[str] = patch.splitlines(keepends = True)
    patched_text: str = ''

    pidx: int = 0 # Index line into patch
    tidx: int = 0 # Index line into text

    # Determine whether to apply changes or revert them
    ## midx is the index into the header to determine what to apply
    ## sign is whether to consider the addition or removal as the final transformer
    midx, sign = (1,'+') if not revert else (3,'-') # midx, sign: int, str

    # Skip header lines
    while pidx < len(patch) and patch[pidx].startswith(("---","+++")):
        pidx += 1

    # Apply patches as long as there are still patch lines left
    while pidx < len(patch):
        # Get header
        header: Optional[re.Match[str]] = _HUNK_HEADER.match(patch[pidx])
        if not header: # If there is no hunk header, throw an exception
            raise PatchError(f'No header found for new hunk on line {pidx}')

        # Get the first line index in the text file to apply/revert the patch for
        shidx = int(header.group(midx)) - 1 + (header.group(midx + 1) == '0')
        # Make sure start hunk index is not after current text index
        ## and that the start hunk index is not after the number of lines in the text file
        if tidx > shidx:
            raise PatchError(f'Hunk start index {shidx} is after the current text index {tidx}')
        if shidx > len(text):
            raise PatchError(f'Hunk start index {shidx} is after the end of the text')

        # Add any lines before the hunk start index to the result
        patched_text += ''.join(text[tidx:shidx])

        # Set the new text start index and move to next patch line
        tidx = shidx
        pidx += 1

        # Loop through patch until eol or next hunk
        while pidx < len(patch) and patch[pidx][0] != '@':
            # If the next patch line is not the end of the text
            ## and the first character isn't a backslash
            ### (indicates either \t, \n, or the _NO_EOL text)
            if pidx + 1 < len(patch) and patch[pidx + 1][0] == '\\':
                # Get the line without the last character
                ## and increase the patch index by 2
                line: str = patch[pidx][:-1]
                pidx += 2
            else:
                # Otherwise, get the current line
                line: str = patch[pidx]
                pidx += 1
            # If the line isn't empty
            if len(line) > 0:
                # If the first character of the line is the sign
                ## or a space
                if line[0] == sign or line[0] == ' ':
                    # Apply the patched text without the first character
                    patched_text += line[1:]
                # Skip the line if it is not the current sign being checked for
                tidx += (line[0] != sign)
    # Apply the rest of the text after all patches were made
    patched_text += ''.join(text[tidx:])
    return patched_text



"""A script containing the necessary methods for creating
and generating patches for a project.
"""

import os
from pathlib import PurePath
import shutil
import json
from datetime import datetime
from typing import Optional
from urllib.parse import urlparse
from requests import Response
from project_patcher.singleton import METADATA_CODEC, METADATA_BUILDER
from project_patcher.utils import download_file
from project_patcher.metadata.base import ProjectMetadata
from project_patcher.workspace.patcher import apply_patch, create_patch

_PROJECT_METADATA_NAME: str = 'project_metadata.json'
"""The file name of the project metadata."""

_PATCH_EXTENSION: str = 'patch'
"""The extension of a patch file."""

_TMP_DIR: str = '.tmp'
"""The directory for temporary files or directories."""

# TODO: Expand to read metadata inside another json object
def read_metadata(dirpath: str = os.curdir, import_loc: Optional[str] = None) -> ProjectMetadata:
    """Creates or reads project metadata for the current / to-be workspace.

    Parameters
    ----------
    dirpath : str (default '.')
        The directory of the (to-be) workspace.
    import_loc: str | None (default None)
        The location to read the project metadata from. Either a directory or a url.
    
    Returns
    -------
    project_patcher.metadata.base.ProjectMetadata
        The metadata for the current / to-be workspace.
    """

    # First check if an import is specified
    if import_loc is not None:
        # Check if import is a url
        if urlparse(import_loc).scheme in ('http', 'https'):
            json_bytes: bytes = None

            def get_metadata_str(response: Response) -> bool:
                """Reads the bytes of a response object into a higher scope variable.
                
                Parameters
                ----------
                response: response.Response
                    The response of the request.
                
                Returns
                -------
                bool
                    Whether the operation was successfully executed.
                """

                nonlocal json_bytes
                json_bytes = response.content
                return True

            # Download file
            if download_file(import_loc,
                lambda response, _: get_metadata_str(response),
                stream = False
            ):
                return write_metadata_to_file(dirpath,
                    METADATA_CODEC.decode(json.loads(json_bytes)))

        # Otherwise, assume import is a path and check if it exists
        elif os.path.exists(import_loc):
            return write_metadata_to_file(dirpath, read_metadata_from_file(import_loc))

    # If none, check if project metadata exists in directory
    if os.path.exists((path := os.sep.join([dirpath, _PROJECT_METADATA_NAME]))):
        return read_metadata_from_file(path)

    # Otherwise, open the builder
    return write_metadata_to_file(dirpath, METADATA_BUILDER())

def read_metadata_from_file(path: str) -> ProjectMetadata:
    """Reads a project metadata from a file location.

    Parameters
    ----------
    path : str
        The path to the project metadata JSON.

    Returns
    -------
    project_patcher.metadata.base.ProjectMetadata
        The metadata for the current / to-be workspace.
    """

    with open(path, mode = 'r', encoding = 'UTF-8') as file:
        return METADATA_CODEC.decode(json.load(file))

def write_metadata_to_file(dirpath: str, metadata: ProjectMetadata) -> ProjectMetadata:
    """Writes a project metadata, named `project_metadata.json`, to a file location.

    Parameters
    ----------
    dirpath : str
        The directory to write the project metadata to.
    metadata : project_patcher.metadata.base.ProjectMetadata
        The metadata for the current workspace.
    
    Returns
    -------
    metadata : project_patcher.metadata.base.ProjectMetadata
        The metadata for the current workspace.
    """

    with open(os.sep.join([dirpath, _PROJECT_METADATA_NAME]),
            mode = 'w', encoding = 'UTF-8') as file:
        print(json.dumps(METADATA_CODEC.encode(metadata), indent = 4), file = file)

    return metadata

def setup_clean(metadata: ProjectMetadata, clean_dir: str = '_clean',
        invalidate_cache: bool = False) -> bool:
    """Generates a clean workspace from the project metadata.

    Parameters
    ----------
    metadata : project_patcher.metadata.base.ProjectMetadata
        The metadata for the current workspace.
    clean_dir : str (default '_clean')
        The directory to generate the clean workspace within.
    invalidate_cache : bool (default False)
        When `True`, removes any cached files from the clean workspace.
    
    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # If the cache should be invalidated, delete the clean directory
    if invalidate_cache and os.path.exists(clean_dir) and os.path.isdir(clean_dir):
        shutil.rmtree(clean_dir)

    # If the cache exists, then skip generation
    ## Otherwise generate the metadata information
    return True if os.path.exists(clean_dir) and os.path.isdir(clean_dir) \
        else metadata.setup(clean_dir)

def apply_patches(working_dir: str = '_src', patch_dir: str = '_patches') -> bool:
    """Applies patches to the working directory.

    Parameters
    ----------
    working_dir : str (default '_src')
        The directory containing the clean workspace to-be patched. 
    patch_dir : str (default '_patches')
        The directory containing the patches for the project files.
    
    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # Assume both directories are present
    for subdir, _, files in os.walk(patch_dir):
        for file in files:
            patch_path: str = os.path.join(subdir, file)
            # Get the relative path of the file for the working directory
            rel_path: str = patch_path[(len(patch_dir) + 1):-(len(_PATCH_EXTENSION) + 1)]

            # Apply patch to working directory
            with open(patch_path, mode = 'r', encoding = 'UTF-8') as patch_file, \
                    open(os.path.join(working_dir, rel_path),
                        mode = 'r+', encoding = 'UTF-8') as work_file:
                work_patch: str = apply_patch(work_file.read(), patch_file.read())
                # Update work file with new information
                work_file.seek(0)
                work_file.write(work_patch)
                work_file.truncate()

    return True

def setup_working(clean_dir: str = '_clean', working_dir: str = '_src',
        patch_dir: str = '_patches', out_dir: str = '_out', include_hidden: bool = False) -> bool:
    """Generates a working directory from the project metadata and any additional
    files and patches.

    Parameters
    ----------
    clean_dir : str (default '_clean')
        The directory containing the raw project files.
    working_dir : str (default '_src')
        The directory containing the clean workspace to-be patched. 
    patch_dir : str (default '_patches')
        The directory containing the patches for the project files.
    out_dir : str (default '_out')
        The directory containing additional files for the workspace.
    include_hidden : bool (default False)
        When `True`, copies hidden files to the working directory.
    
    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # Remove existing working directory if exists
    if os.path.exists(working_dir) and os.path.isdir(working_dir):
        shutil.rmtree(working_dir)

    # Generate working directory (shouldn't exist)
    os.makedirs(working_dir)

    # Copy clean directory into working directory (clean directory must exist)
    if include_hidden:
        shutil.copytree(clean_dir, working_dir, dirs_exist_ok = True)
    else:
        shutil.copytree(clean_dir, working_dir, dirs_exist_ok = True,
            ignore = shutil.ignore_patterns('.*'))

    return setup_working_raw(working_dir = working_dir, patch_dir = patch_dir,
        out_dir = out_dir)

def setup_working_raw(working_dir: str = '_src', patch_dir: str = '_patches',
        out_dir: str = '_out') -> bool:
    """Generates a working directory from the project metadata and any additional
    files and patches. Performs no validation checks.

    Parameters
    ----------
    working_dir : str (default '_src')
        The directory containing the clean workspace to-be patched. 
    patch_dir : str (default '_patches')
        The directory containing the patches for the project files.
    out_dir : str (default '_out')
        The directory containing additional files for the workspace.
    
    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # If an output directory exists, copy into working directory
    if os.path.exists(out_dir) and os.path.isdir(out_dir):
        shutil.copytree(out_dir, working_dir, dirs_exist_ok = True)

    # If the patches directory exists, apply patches to working directory
    return apply_patches(working_dir, patch_dir) \
        if os.path.exists(patch_dir) and os.path.isdir(patch_dir) \
        else True

def check_existing_patch(rel_patch_path: str, patch_path: str,
        patch_text: str, patch_dir: str = '_patches') -> bool:
    """Returns whether there is an equivalent, existing patch and copies it from the
    temporary directory.

    Parameters
    ----------
    rel_patch_path : str
        The relative path to the patch.
    patch_path : str
        The path to the patch output location.
    patch_text : str
        The text of the patch file.
    patch_dir : str (default '_patches')
        The directory containing the patches for the project files.

    Returns
    -------
    bool
        If `True`, there is an equivalent, existing patch
    """

    if os.path.exists(temp_patch_path :=
            os.path.join(os.path.join(_TMP_DIR, patch_dir), rel_patch_path)):
        # Read existing patch for comparison
        patch_text_no_head: str = ''.join(patch_text.splitlines(keepends = True)[2:])
        temp_patch_text: Optional[str] = None
        with open(temp_patch_path, mode = 'r', encoding = 'UTF-8') as temp_patch_file:
            temp_patch_text: str = ''.join(temp_patch_file.readlines()[2:])

        # If patches are equivalent, move file
        if patch_text_no_head == temp_patch_text:
            shutil.move(temp_patch_path, patch_path)
            return True

    return False

def generate_patch(path: str, work_path: str, clean_path: str,
        patch_dir: str = '_patches', time: str = str(datetime.now())) -> bool:
    """Generates a patch between two files if they are not equal.

    Parameters
    ----------
    path : str
        The relative path of the file.
    work_path : str
        The path of the file in the working directory.
    clean_path : str
        The path of the file in the clean directory.
    patch_dir : str (default '_patches')
        The directory containing the patches for the project files.
    time : str  (default `datetime.datetime.now`)
        The time the patch was generated.
    
    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # Assume patches directory exists

    with open(work_path, mode = 'r', encoding = 'UTF-8') as work_file, \
            open(clean_path, mode = 'r', encoding = 'UTF-8') as clean_file:
        # Generate patch file if not empty
        if (patch_text := create_patch(clean_file.read(), work_file.read(),
                filename = path,
                time = time)):
            rel_patch_path: str = os.extsep.join([path, _PATCH_EXTENSION])
            patch_path: str = os.path.join(patch_dir, rel_patch_path)

            # Create directory if necessary
            os.makedirs(os.path.dirname(patch_path), exist_ok = True)

            if check_existing_patch(rel_patch_path, patch_path, patch_text, patch_dir = patch_dir):
                return True

            # Otherwise write new patch
            with open(patch_path, mode = 'w', encoding = 'UTF-8') as patch_file:
                patch_file.write(patch_text)

    return True

def output_file(path: str, work_path: str, out_dir: str = '_out') -> bool:
    """Copies an additional file for the workspace to the output directory.
    
    Parameters
    ----------
    path : str
        The relative path of the file.
    work_path : str
        The path of the file in the working directory.
    out_dir : str (default '_out')
        The directory containing additional files for the workspace.

    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # Assume output directory exists

    out_path: str = os.path.join(out_dir, path)
    # Create directory if necessary
    os.makedirs(os.path.dirname(out_path), exist_ok = True)
    shutil.copy(work_path, out_path)

    return True

def output_working(metadata: ProjectMetadata, clean_dir: str = '_clean', working_dir: str = '_src',
        patch_dir: str = '_patches', out_dir: str = '_out') -> bool:
    """Generates the patches and copies any additional files for construction.
    
    Parameters
    ----------
    metadata : project_patcher.metadata.base.ProjectMetadata
        The metadata for the current workspace.
    clean_dir : str (default '_clean')
        The directory containing the raw project files.
    working_dir : str (default '_src')
        The directory containing the clean workspace to-be patched. 
    patch_dir : str (default '_patches')
        The directory containing the patches for the project files.
    out_dir : str (default '_out')
        The directory containing additional files for the workspace.

    Returns
    -------
    bool
        Whether the operation was successfully executed.
    """

    # Basic variables
    time: str = str(datetime.now())

    # Create temp dir (shouldn't exist)
    os.makedirs(_TMP_DIR, exist_ok = True)

    # If patch directory and output exist, delete them
    if os.path.exists(out_dir) and os.path.isdir(out_dir):
        shutil.rmtree(out_dir)
    if os.path.exists(patch_dir) and os.path.isdir(patch_dir):
        shutil.move(patch_dir, _TMP_DIR)

    # Generate ignored and overwritten list
    ignore, overwrite = metadata.ignore_and_overwrite(working_dir) # Set[str], Set[str]

    for subdir, _, files in os.walk(working_dir):
        for file in files:
            # Setup paths
            work_path: str = os.path.join(subdir, file)
            rel_path: str = work_path[(len(working_dir) + 1):]
            clean_path: str = os.path.join(clean_dir, rel_path)
            rel_path_posix: str = PurePath(rel_path).as_posix()

            # If clean file exists, generate patch and write
            if os.path.exists(clean_path):
                if rel_path_posix in ignore:
                    pass # Ignore patch generation if in relative
                elif rel_path_posix in overwrite:
                    # Copy file to output if overwrite
                    output_file(rel_path, work_path, out_dir = out_dir)
                else:
                    # Otherwise generate the patch
                    generate_patch(rel_path, work_path, clean_path,
                        patch_dir = patch_dir, time = time)

            # Otherwise output files to directory
            else:
                output_file(rel_path, work_path, out_dir = out_dir)

    # Delete temp directory afterwards
    shutil.rmtree(_TMP_DIR)
    return True



"""A script dedicated to setting up any registries or singleton
objects used within this package.
"""

from typing import Set, Callable
from project_patcher.struct.registry import Registry
from project_patcher.metadata.file import ProjectFileCodec, ProjectFile
from project_patcher.metadata.base import ProjectMetadataCodec, ProjectMetadata, build_metadata
from project_patcher.utils import has_module
from project_patcher.metadata.files.osf import OSFProjectFileCodec, build_osf
from project_patcher.metadata.files.url import URLProjectFileCodec, build_url

# Add registries

PROJECT_FILE_TYPES: Registry['ProjectFileCodec'] = Registry()
"""A registry containing the codecs for :class:`project_patcher.metadata.file.ProjectFileCodec`s."""

PROJECT_FILE_BUILDERS: Registry[Callable[[], ProjectFile]] = Registry()
"""A registry containing the builders for :class:`project_patcher.metadata.file.ProjectFile`s."""

OPTIONAL_DEPENDENCIES: Set[str] = set()
"""A set containing the loaded optional dependencies."""

# Register optional dependencies

if has_module('git'):
    OPTIONAL_DEPENDENCIES.add('git')

# Register Codecs

OSF_FILE_CODEC: OSFProjectFileCodec = OSFProjectFileCodec()
"""The codec for :class:`project_patcher.metadata.files.osf.OSFProjectFile`s."""
PROJECT_FILE_TYPES['osf'] = OSF_FILE_CODEC
PROJECT_FILE_BUILDERS['osf'] = build_osf

URL_FILE_CODEC: URLProjectFileCodec = URLProjectFileCodec()
"""The codec for :class:`project_patcher.metadata.files.url.URLProjectFile`s."""
PROJECT_FILE_TYPES['url'] = URL_FILE_CODEC
PROJECT_FILE_BUILDERS['url'] = build_url

if 'git' in OPTIONAL_DEPENDENCIES:
    from project_patcher.metadata.files.gitrepo import GitProjectFileCodec, build_git

    GIT_FILE_CODEC: GitProjectFileCodec = GitProjectFileCodec()
    """The codec for :class:`project_patcher.metadata.files.git.GitProjectFile`s."""

    PROJECT_FILE_TYPES['git'] = GIT_FILE_CODEC
    PROJECT_FILE_BUILDERS['git'] = build_git

METADATA_CODEC: 'ProjectMetadataCodec' = ProjectMetadataCodec()
"""The codec for :class:`project_patcher.metadata.base.ProjectMetadata`."""
METADATA_BUILDER: Callable[[], ProjectMetadata] = build_metadata
"""The builder for a :class:`project_patcher.metadata.base.ProjectMetadata`."""



"""A script containing a generic implementation of an encoder and decoder
for an object and a dictionary.
"""

from abc import ABC, abstractmethod
from typing import TypeVar, Type, Dict, Any, Generic

DataObject = TypeVar('DataObject')
"""The type of the object holding data."""

DictObject: Type[Dict[str, Any]] = Dict[str, Any]
"""A generic intermediate object for reading or writing to a file."""

class DictCodec(ABC, Generic[DataObject]):
    """An abstract, generic encoder and decoder between a dictionary and a data object.

    Types
    -----
    DataObject
        The type of the data object to be encoded or decoded to.
    """

    @abstractmethod
    def decode(self, obj: DictObject) -> DataObject:
        """Decodes a dictionary to the data object.
        
        Parameters
        ----------
        obj : Dict[str, Any]
            The dictionary containing the data for the object.
        
        Returns
        -------
        DataObject
            The decoded data object.
        """

    @abstractmethod
    def encode(self, obj: DataObject) -> DictObject:
        """Encodes a data object to a dictionary.
        
        Parameters
        ----------
        obj : DataObject
            The object containing the data.
        
        Returns
        -------
        Dict[str, Any]
            The encoded data object in a dictionary.
        """



"""A script containing the methods needed for command line integration.
"""

from typing import Optional
import click
import project_patcher.workspace.project as wspc
from project_patcher.metadata.base import ProjectMetadata

@click.group()
def main() -> None:
    """A command line interface to construct projects,
    generated diffs, and patch multiple files in one
    implementation.
    """

@main.command(name = 'init')
@click.option(
    '--import_metadata', '-I',
    type = str,
    default = None,
    help = 'A path or URL to the metadata JSON.'
)
@click.option(
    '-a', '-A', 'include_hidden',
    is_flag = True,
    help = 'When added, copies hidden files to the working directory.'
)
def init(import_metadata: Optional[str] = None, include_hidden: bool = False) -> None:
    """Initializes a new project or an existing project from the
    metadata JSON in the executing directory, an import, or from
    the metadata builder if neither are present.
    """

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(import_loc = import_metadata)

    # Setup workspace
    wspc.setup_clean(metadata)
    wspc.setup_working(include_hidden = include_hidden)

    print('Initialized patching environment!')

@main.command(name = 'output')
def output() -> None:
    """Generates any patches and clones the new files to an output
    directory."""

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata()

    # Output working and generate patches
    wspc.output_working(metadata)

    print('Generated patches and output files!')

@main.command(name = 'clean')
@click.option(
    '--import_metadata', '-I',
    type = str,
    default = None,
    help = 'A path or URL to the metadata JSON.'
)
def clean(import_metadata: Optional[str] = None) -> None:
    """Initializes a clean workspace from the
    metadata JSON in the executing directory, an import, or from
    the metadata builder if neither are present.
    """

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(import_loc = import_metadata)

    # Setup workspace
    wspc.setup_clean(metadata)

    print('Setup clean workspace!')

@main.command(name = 'src')
@click.option(
    '--import_metadata', '-I',
    type = str,
    default = None,
    help = 'A path or URL to the metadata JSON.'
)
def source(import_metadata: Optional[str] = None) -> None:
    """Initializes a patched workspace from the
    metadata JSON in the executing directory, an import, or from
    the metadata builder if neither are present.
    """

    # Get metadata
    metadata: ProjectMetadata = wspc.read_metadata(import_loc = import_metadata)

    # Setup workspace
    wspc.setup_clean(metadata, '_src', invalidate_cache = True)
    wspc.setup_working_raw()

    print('Setup patched workspace!')
